from typing import Callable, Optional, Iterable, Tuple, List
from .internals.threaded_reader import threaded_reader
from .internals.experimental_processed_reader import processed_reader
from .internals.parsers import pass_thru_parser, block_parser, json_parser
from .internals.filters import Filters
from ..formats.dictset import select_record_fields, select_from
from ..formats.dictset.display import html_table, ascii_table
from ..formats import json
from ...logging import get_logger
from ...errors import InvalidCombinationError

# available parsers
PARSERS = {
    "json": json_parser,
    "text": pass_thru_parser,
    "block": block_parser,
    "pass-thru": pass_thru_parser
}


class Reader():

    def __init__(
        self,
        *,  # force all paramters to be keyworded
        select: list = ['*'],
        dataset: str = None,
        where: Optional[Callable] = None,
        filters: Optional[List[Tuple[str, str, object]]] = None,
        inner_reader = None,   # type:ignore
        row_format: str = "json",
        **kwargs):
        """
        Reads records from a data store, opinionated toward Google Cloud Storage
        but a filesystem reader is available to assist with local development.

        The Reader will iterate over a set of files and return them to the caller
        as a single stream of records. The files can be read from a single folder
        or can be matched over a set of date/time formatted folder names. This is
        useful to read over a set of logs. The date range is provided as part of 
        the call; this is essentially a way to partition the data by date/time.

        The reader can filter records to return a subset, for JSON formatted data
        the records can be converted to dictionaries before filtering. JSON data
        can also be used to select columns, so not all read data is returned.

        The reader does not support aggregations, calculations or grouping of data,
        it is a log reader and returns log entries. The reader can convert a set
        into _Pandas_ dataframe, or the _dictset_ helper library can perform some 
        activities on the set in a more memory efficient manner.

        Note:
            Different _inner_readers_ may take or require additional parameters.

        Parameters:
            select: list of strings (optional):
                A list of the names of the columns to return from the dataset,
                the default is all columns
            dataset: string:
                The path to the data
            where: callable (optional):
                **TO BE DEPRECATED**
                A method (function or lambda expression) to filter the returned
                records, where the function returns True the record is
                returned, False the record is skipped. The default is all
                records
            filters: List of tuples (optional)
                **EXPERIMENTAL**
                Rows which do not match the filter predicate will be removed
                from scanned data. Default is no filtering.
                Each tuple has format: (`key`, `op`, `value`) and compares the
                key with the value. The supported op are: `=` or `==`, `!=`, 
                `<`, `>`, `<=`, `>=`, `in`, `!in` (not in) and `like`. If the 
                `op` is `in` or `!in`, the `value` must be a collection such as
                a _list_, a _set_ or a _tuple_.
                `like` performs similar to the SQL operator `%` is a
                multicharacter wildcard and `_` is a single character wildcard.
            inner_reader: BaseReader (optional):
                The reader class to perform the data access Operators, the
                default is GoogleCloudStorageReader
            row_format: string (optional):
                Controls how the data is interpretted. 'json' will parse to a
                dictionary before _select_ or _where_ is applied, 'text' will 
                just return the line that has been read, 'block' will return
                the content of a file as a record. the default is 'json'.
            start_date: datetime (optional):
                The starting date of the range to read over - if used with
                _date_range_, this value will be preferred, default is today
            end_date: datetime (optional):
                The end date of the range to read over - if used with
                _date_range_, this value will be preferred, default is today
            thread_count: integer (optional):
                Use multiple threads to read data files, the default is to not
                use additional threads, the maximum number of threads is 8
            step_back_days: integer (optional):
                The number of days to look back if data for the current date is
                not available.
            fork_processes: boolean (experimental):
                **EXPERIMENTAL**
                Create parallel processes to read data files

        Yields:
            dictionary (string if data format is 'text')

        Raises:
            TypeError
                Reader _select_ parameter must be a list
            TypeError
                Reader _where_ parameter must be Callable or None
            TypeError
                Data format unsupported
            InvalidCombinationError
                Forking and Threading can not be used at the same time
        """
        if not isinstance(select, list):  # pragma: no cover
            raise TypeError("Reader 'select' parameter must be a list")
        if where is not None and not hasattr(where, '__call__'):
            raise TypeError("Reader 'where' parameter must be Callable or None")

        # load the line converter
        self._parse = PARSERS.get(row_format.lower())
        if self._parse is None:  # pragma: no cover
            raise TypeError(F"Row format unsupported: {row_format} - valid options are {list(PARSERS.keys())}.")

        if dataset is None and kwargs.get('from_path'):  # pragma: no cover
            dataset = kwargs['from_path']
            get_logger().warning('DEPRECATION: Reader \'from_path\' parameter will be replaced with \'dataset\' ')

        # lazy loading of dependency
        if inner_reader is None:
            from ...adapters.google import GoogleCloudStorageReader
            inner_reader = GoogleCloudStorageReader
        # instantiate the injected reader class
        self.reader_class = inner_reader(dataset=dataset, **kwargs)  # type:ignore

        self.select = select.copy()
        self.where: Optional[Callable] = where

        # initialize the reader
        self._inner_line_reader = None

        arg_dict = kwargs.copy()
        arg_dict['select'] = F'{select}'
        arg_dict['dataset'] = F'{dataset}'
        arg_dict['where'] = F'{where.__name__ if not where is None else "Select All"}'
        arg_dict['inner_reader'] = F'{inner_reader.__name__}'  # type:ignore
        arg_dict['row_format'] = F'{row_format}'
        get_logger().debug(json.serialize(arg_dict))

        # threaded reader
        self.thread_count = int(kwargs.get('thread_count', 0))

        # number of days to walk backwards to find records
        self.step_back_days = int(kwargs.get('step_back_days', 0))
        if self.step_back_days > 0 and self.reader_class.start_date != self.reader_class.end_date:  # pragma: no cover
            raise InvalidCombinationError("step_back_days can only be used when the start and end dates are the same")

        if row_format != 'pass-thru' and kwargs.get('extension') == '.parquet':  # pragma: no cover
            raise InvalidCombinationError("`parquet` extension much be used with the `pass-thru` row_format")

        """ FEATURES IN DEVELOPMENT """

        # multiprocessed reader
        self.fork_processes = bool(kwargs.get('fork_processes', False))
        if self.thread_count > 0 and self.fork_processes:  # pragma: no cover
            raise InvalidCombinationError('Forking and Threading can not be used at the same time')
        if self.fork_processes:
            get_logger().warning("FORKED READER IS EXPERIMENTAL")

        self.filters = None
        if filters:
            self.filters = Filters(filters)   
            get_logger().warning("FILTERS IS EXPERIMENTAL")
            if where:
                raise InvalidCombinationError('Where and Filters can not be used at the same time')
        if where:
            get_logger().warning("`where` is a deprecation target, use `filters` or `dictset.select_from` instead")
        
    """
    Iterable

    Use this class as an iterable:

        for line in Reader("file"):
            print(line)
    """

    def _create_line_reader(self):
        blob_list = self.reader_class.get_list_of_blobs()

        # handle stepping back if the option is set
        if self.step_back_days > 0:
            while not bool(blob_list) and self.step_back_days >= self.reader_class.days_stepped_back:
                self.reader_class.step_back_a_day()
                blob_list = self.reader_class.get_list_of_blobs()
            if self.step_back_days < self.reader_class.days_stepped_back:
                get_logger().alert(F"No data found in last {self.step_back_days} days - aborting")
                import sys
                sys.exit(-1) 
            if self.reader_class.days_stepped_back > 0:
                get_logger().warning(F"Stepped back {self.reader_class.days_stepped_back} days to {self.reader_class.start_date} to find last data, my limit is {self.step_back_days} days.")

        get_logger().debug(F"Reader found {len(blob_list)} sources to read data from.")
        
        if self.thread_count > 0:
            ds = threaded_reader(blob_list, self.reader_class, self.thread_count)
            ds = self._parse(ds)

            if self.filters:
                yield from self.filters.filter_dictset(ds)
            else:
                yield from select_from(ds, where=self.where)
        elif self.fork_processes:
            yield from processed_reader(list(blob_list), self.reader_class, self._parse, self.where)
        else:
            for blob in blob_list:
                get_logger().debug(F"Reading from {blob}")
                ds = self.reader_class.get_records(blob)
                ds = self._parse(ds)
                if self.filters:
                    yield from self.filters.filter_dictset(ds)
                else:
                    yield from select_from(ds, where=self.where)

    def __iter__(self):
        return self

    def __next__(self):
        if self._inner_line_reader is None:
            self._inner_line_reader = self._create_line_reader()

        # get the the next line from the reader
        record = self._inner_line_reader.__next__()
        if self.select != ['*']:
            record = select_record_fields(record, self.select)
        return record


    """
    Context Manager

    Use this class using the 'with' statement:

        with Reader("file") as r:
            line = r.read_line()
            while line:
                print(line)
    """

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass  # exist needs to exist to be a context manager

    def read_line(self):
        """
        Read the next line from the _Reader_.

        Returns:
            dictionary (or string)
        """
        try:
            return self.__next__()
        except StopIteration:
            return None

    """
    Exports
    """
    def to_pandas(self):
        """
        Load the contents of the _Reader_ to a _Pandas_ DataFrame.

        Returns:
            Pandas DataFrame
        """
        try:
            import pandas as pd  # type:ignore
        except ImportError:  # pragma: no cover
            raise ImportError("Pandas must be installed to use 'to_pandas'")
        return pd.DataFrame(self)

    def __repr__(self):  # pragma: no cover

        def is_running_from_ipython():
            """
            True when running in Jupyter
            """
            try:
                from IPython import get_ipython  # type:ignore
                return get_ipython() is not None
            except Exception:
                return False

        if is_running_from_ipython():
            from IPython.display import HTML, display  # type:ignore
            html = html_table(self, 5)
            display(HTML(html))
            return ''  # __repr__ must return something
        else:
            return ascii_table(self, 5)



"""



Adapted from http://pts.github.io/pts-line-bisect/line_bisect_evolution.html

License: GNU GPL v2 or newer, at your choice.
Accessed: 2021-03-06
"""


import orjson as json
from functools import lru_cache


@lru_cache(8)
def _read_and_compare(ofs, f, size, tester):
    if ofs:
        if f.tell() != ofs - 1:  # Avoid lseek(2) call if not needed.
            f.seek(ofs - 1)  # Just to figure out where our line starts.
        f.readline()  # Ignore previous line, find our line.
        fofs = min(size, f.tell())
    else:
        fofs = 0
    g = True
    if fofs < size:
        if not fofs and f.tell():
            f.seek(0)
        line = f.readline()
        line = json.loads(line)
        line = line['key']
        g = tester <= line
    return [fofs, g, ofs]

def bisect_way(f, x):
    f.seek(0, 2)
    size = f.tell()
    lo, hi, mid = 0, size - 1, 1
    while lo < hi:
        mid = (lo + hi) >> 1
        midf, g, _ = _read_and_compare(mid, f, size, x)
        if g:
            hi = mid
        else:
            lo = mid + 1
        if mid != lo:
            midf = _read_and_compare(lo, f, size, x)[0]
    return midf



import sys
import time

def binary(value):
    with open('sorted_twitter.index', 'r') as f:
        start = bisect_way(f, value)
        f.seek(start)
        data = f.readline()
        return data
        #sys.stdout.write(data + '\n')


value = '~~kn~~'

t = time.time_ns()
for rep in range(500):
    r = binary(value)
print('binary', (time.time_ns() - t) / 1e9, r)



import os
import re
import hashlib
import logging
import json

# import ujson because importing the mabel json module creates circular
# references, we use ujson rather than orjson for greatest compatibility


# if we find a key which matches these strings, we hash the contents
KEYS_TO_SANITIZE = ["password$", "pwd$", ".*_secret$", ".*_key$"]
COLOR_EXCHANGES = {
    ' ALERT    ': '\033[1;31m' + ' ALERT    ' + '\033[0m',  # BOLD RED
    ' ERROR    ': '\033[31m' + ' ERROR    ' + '\033[0m',    # RED 
    ' DEBUG    ': '\033[32m' + ' DEBUG    ' + '\033[0m',    # GREEN
    ' AUDIT    ': '\033[33m' + ' AUDIT    ' + '\033[0m',    # YELLOW
    ' WARNING  ': '\033[36m' + ' WARNING  ' + '\033[0m',    # CYAN
}

class SanitizingLogFormatter(logging.Formatter):

    def __init__(self, orig_formatter):
        """
        Remove sensitive data from records before saving to external logs.
        Note that the value is hashed using (SHA256) and only the first 8 
        characters of the hex-encoded hash are presented. This information
        allows values to be traced without disclosing the actual value. 

        The Sanitizer can only sanitize dictionaries, it doesn't
        sanitize strings, which could contain sensitive information
        We use the message id as a salt to further protect sensitive 
        information.

        Based On: https://github.com/joocer/cronicl/blob/main/cronicl/utils/sanitizer.py
        """
        self.orig_formatter = orig_formatter

    def format(self, record):
        msg = self.orig_formatter.format(record)
        msg = self.sanitize_record(msg)
        if '://' in msg:
            msg = re.sub(r':\/\/(.*?)\@', r'://<redacted>', msg)
        return msg
    
    def _can_colorize(self):
        colorterm = os.environ.get('COLORTERM', '').lower()
        term = os.environ.get('TERM', '').lower()
        return 'yes' in colorterm or 'true' in colorterm or '256' in term        

    def color_code(self, record):
        if self._can_colorize():
            for k, v in COLOR_EXCHANGES.items():
                if k in record:
                    return record.replace(k, v)
        return record

    def colorize(self, record):
        #[MBL-70]
        # https://gist.github.com/vratiu/9780109
        return record

    def __getattr__(self, attr):
        return getattr(self.orig_formatter, attr)

    def hash_it(self, value_to_hash):
        return hashlib.sha256(value_to_hash.encode()).hexdigest()[:8]

    def sanitize_record(self, record):

        record = self.color_code(record)
        parts = record.split('|')
        json_part = parts.pop()
        done_anything = False

        try:
            dirty_record = json.loads(json_part)
        except:
            return record

        if isinstance(dirty_record, dict):
            for key, value in dirty_record.items():
                if any([True for expression in KEYS_TO_SANITIZE if re.match(expression, key, re.IGNORECASE)]):
                    dirty_record[key] = '<redacted:' + self.hash_it(str(value)) + '>'
                    done_anything = True

        if done_anything:
            parts.append(' ' + json.dumps(dirty_record))
            return '|'.join(parts)

        return record



"""
Null Writer

Impotent writer for testing, writes to the log to help with debugging.

#nodoc - don't add to the documentation wiki
"""
from ...logging import get_logger
from ...utils import paths
from ...data.writers.internals.base_inner_writer import BaseInnerWriter


class NullWriter(BaseInnerWriter):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        kwargs_passed = [f"{k}={v!r}" for k, v in kwargs.items()]
        self.formatted_args = ", ".join(kwargs_passed)

    def commit(
            self,
            byte_data,
            override_blob_name=None):

        # if we've been given the filename, use that, otherwise get the
        # name from the path builder
        if override_blob_name:
            blob_name = override_blob_name
        else:
            blob_name = self._build_path()

        return F"NullWriter({blob_name})"



"""
MinIo Reader - may work with AWS
"""
import io
from ...utils import paths, common
from ...data.readers.internals.base_inner_reader import BaseInnerReader
try:
    from minio import Minio  # type:ignore
except ImportError:  # pragma: no cover
    pass


class MinIoReader(BaseInnerReader):

    def __init__(
            self,
            end_point: str,
            access_key: str,
            secret_key: str,
            **kwargs):
        super().__init__(**kwargs)
        secure = kwargs.get('secure', True)
        self.minio = Minio(end_point, access_key, secret_key, secure=secure)

    def get_blobs_at_path(self, path):
        bucket, object_path, _, _ = paths.get_parts(path)
        for cycle_date in common.date_range(self.start_date, self.end_date):
            cycle_path = paths.build_path(path=object_path, date=cycle_date)
            blobs = self.minio.list_objects(
                    bucket_name=bucket,
                    prefix=cycle_path,
                    recursive=True)

            yield from [bucket + '/' + blob.object_name
                    for blob in blobs
                    if not blob.object_name.endswith('/')]

    def get_blob_stream(self, blob_name: str) -> io.IOBase:
        bucket, object_path, name, extension = paths.get_parts(blob_name)
        stream = self.minio.get_object(bucket, object_path + name + extension).read()

        io_stream = io.BytesIO(stream)
        return io_stream



"""
The WriterPool maintains multiple BlobWriters.

It allows new ParitionWriters be be added to the pool, and when the size of the
pool exceeds the stated size limit, it can recommend the least recently used
BlobWriters to be evicted. Note the pool_size is not a hard limit, there
can be more active BlobWriters, this is used to determine how many
BlobWriters to recommend for evict.
"""
import time
import threading
from .blob_writer import BlobWriter
from ...formats.dictset import display


class WriterPool():

    __slots__ = ('writers', 'pool_size', 'kwargs')

    def __init__(
            self,
            pool_size: int = 5,
            **kwargs):

        self.writers: list = []
        self.pool_size = pool_size
        self.kwargs = kwargs.copy()
        # remove this from the kw args if it's there
        self.kwargs.pop('dataset', '')

    def get_writer(self, identity):
        try:
            # if a writer with the specified identity exists then return it
            # update the last_access status so we can determin the LRU
            writer = [writer for writer in self.writers if writer['identity'] == identity].pop()
            writer['last_access'] = time.time()
            return writer['writer']
        except IndexError:
            # if no writer was found, create a new one and add to the pool.
            # we don't check if we're overflowing before doing this,
            # pool_size is managed by the background thread in the Writer
            writer = self.create_writer(identity)
            self.writers.append(writer)
            return writer.get('writer')

    def create_writer(self, identity):
        # instantiate a new writer
        new_writer = {
            'identity': identity,
            'last_access': time.time(),
            'writer': BlobWriter(dataset=identity, **self.kwargs)
        }
        return new_writer

    def remove_writer(self, identity):
        # remove a writer from the pool and commit
        with threading.Lock():
            writer = [writer for writer in self.writers if writer.get('identity') == identity].pop()
            self.writers = [writer for writer in self.writers if writer.get('identity') != identity]
        writer.get('writer').commit()

    def close(self):
        # evict everyone from the pool
        with threading.Lock():
            [self.remove_writer(writer.get('identity')) for writer in self.writers]
            self.writers = []

    def nominate_writers_to_evict(self):
        # if there are more than pool_size writers in the pool, get a list of
        # LRU ones that exceed the pool_size
        def _inner_sort_key(row):
            return row.get('last_access')
        evictees = sorted(self.writers, key=_inner_sort_key, reverse=False)[:len(self.writers) - self.pool_size]
        return [writer.get('identity') for writer in evictees]

    def get_stale_writers(self, seconds_since_last_use):
        """ get the identities of writers which haven't been accessed recently """
        now = time.time()
        return [writer.get('identity') for writer in self.writers if now - writer.get('last_access') > seconds_since_last_use]

    def __str__(self):
        # much easier to read
        return display.ascii_table(self.writers)



import threading
import os
from time import sleep
from ..logging import get_logger

can_use_resource_lib = True
try:
    import resource
except ImportError:
    can_use_resource_lib = False
    import psutil  # type:ignore


class ResourceMonitor():
    def __init__(self, frequency=1):
        self.keep_measuring = True
        self.frequency = frequency
        self.max_memory_usage = 0
        if not can_use_resource_lib:
            self.process = psutil.Process(os.getpid())

    def resource_usage(self):
        while self.keep_measuring:
            if can_use_resource_lib:
                memory_usage = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
            else:
                memory_usage = self.process.memory_info()[0]
            self.max_memory_usage = max(self.max_memory_usage, memory_usage)
            get_logger().info({"memory":memory_usage})
            sleep(self.frequency)

        return self.max_memory_usage

    def __del__(self):
        get_logger().debug({"max_memory":self.max_memory_usage})


monitor = ResourceMonitor()
resource_thread = threading.Thread(target=monitor.resource_usage)
resource_thread.name = 'mabel-resource-monitor'
resource_thread.daemon = True
resource_thread.start()



import glob
import datetime
from typing import Optional
from ..data.formats.json import parse, serialize
from ..logging import get_logger


def date_range(start_date: Optional[datetime.date], end_date: Optional[datetime.date]):
    # if dates aren't provided, use today
    if not end_date:
        end_date = datetime.date.today()
    if not start_date:
        start_date = datetime.date.today()

    if end_date < start_date:
        raise ValueError("date_range: end_date must be the same or later than the start_date ")

    for n in range(int((end_date - start_date).days) + 1):
        yield start_date + datetime.timedelta(n)


def build_context(**kwargs: dict):
    """
    Build Context takes an arbitrary dictionary and merges with a dictionary
    which reflects configuration read from a json file.

    It builds start and end dates, either from either source or defaults to
    today.
    """
    def read_config(config_file):
        # read the job configuration
        try:
            file_location = glob.glob('**/' + config_file, recursive=True).pop()
            get_logger().debug(F'Reading configuration from {file_location}.')
            with open(file_location, 'r') as f:
                config = parse(f.read())
            return config
        except IndexError as e:
            raise IndexError(F'Error: {e}, Likely Cause: Config file {config_file} not found')
        except ValueError as e:
            raise ValueError(F'Error: {e}, Likely Cause: Config file {config_file} incorrectly formatted')
        except Exception as e:
            if type(e).__name__ == 'JSONDecodeError':
                raise ValueError(F'Config file {config_file} not valid JSON')
            else:
                raise

    # read the configuration file
    config_file = kwargs.get('config_file', 'config.json')
    config = read_config(config_file=config_file)
    if not config.get('config'):
        config['config'] = {}

    # merge the sources
    context = {**config, **kwargs}

    return context


import struct

RECORD_SIZE = 40

def binary_search(f, target):
    left, right = 0, 99999 #len(nums)-1       # define search space: left, right, and middle

    while left <= right:
        middle = (left + right) >> 1

        f.seek(RECORD_SIZE * middle)
        key_len, key, index = struct.unpack("i 32s i", f.read(RECORD_SIZE))
        key = key.decode()[:key_len]

        if key == target:
            return index
        elif key > target:
            right = middle - 1
        else:
            left = middle + 1
    return -1

def write_index():
    import ujson as json
    by = []
    with open('tests/data/formats/tweets.bin', 'wb') as bi:
        with open('tests/data/formats/tweets.index', 'r') as f:
            for i, row in enumerate(f):
                j = json.loads(row)
                by = struct.pack("i 32s i", len(j['key']), j['key'].encode()[:32], i)
                bi.write(by)
    print(len(by))

#write_index()

import time
value = '~~k~~'

def binbin(value):
    with open('tests/data/formats/tweets.bin', 'rb') as f:
        return binary_search(f, value)

t = time.time_ns()
for rep in range(10000):   # ~20x faster
    r = binbin(value)
print('full binary', (time.time_ns() - t) / 1e9, r)

t = time.time_ns()
with open('tests/data/formats/tweets.bin', 'rb') as f:
    for rep in range(20000):   # ~40x faster
        binary_search(f, value)
print('full binary', (time.time_ns() - t) / 1e9, r)


import os
import re
import datetime
from typing import Any, Union, List
from ..formats.json import parse
from ...errors import ValidationError
from ..formats.dictset.display import ascii_table


DEFAULT_MIN = -9223372036854775808
DEFAULT_MAX = 9223372036854775807
VALID_BOOLEAN_VALUES = {"true", "false", "on", "off", "yes", "no", "0", "1"}
CVE_REGEX = re.compile('cve|CVE-[0-9]{4}-[0-9]{4,}')


def is_boolean(**kwargs):
    def _inner(value: Any) -> bool:
        """boolean"""
        return str(value).lower() in VALID_BOOLEAN_VALUES
    return _inner

def is_cve(**kwargs):
    def _inner(value):
        """cve"""
        return CVE_REGEX.match(str(value))
    return _inner

def is_date(**kwargs):
    DATE_SEPARATORS = {'-', '\\', '/', ':'}
    # date validation at speed is hard, dateutil is great but really slow
    # this is as about as good as validating a string is a date, but isn't
    def _inner(value: Any) -> bool:
        """date"""
        try:
            if type(value).__name__ in ("datetime", "date", "time"):
                return True
            if type(value).__name__ == 'str':
                if not value[4] in DATE_SEPARATORS:
                    return False
                if not value[7] in DATE_SEPARATORS:
                    return False
                if len(value) == 10:
                    # YYYY-MM-DD
                    datetime.date(*map(int, [value[:4], value[5:7], value[8:10]]))
                else:
                    if not value[10] == "T":
                        return False
                    if not value[13] in DATE_SEPARATORS:
                        return False
                    if not value[16] in DATE_SEPARATORS:
                        return False                    
                    # YYYY-MM-DDTHH:MM:SS
                    datetime.datetime(*map(int, [value[:4], value[5:7], value[8:10], value[11:13], value[14:16], value[17:19]]))   # type:ignore
            return True
        except (ValueError, TypeError):
            return False
    return _inner

def is_list(**kwargs):
    def _inner(value: Any) -> bool:
        """list"""
        return isinstance(value, (list, set))
    return _inner

def is_null(**kwargs):
    def _inner(value: Any) -> bool:
        """nullable"""
        return (value is None) or (value == '') or (value == [])
    return _inner

def is_numeric(**kwargs):
    mn = kwargs.get('min') or DEFAULT_MIN
    mx = kwargs.get('max') or DEFAULT_MAX
    def _inner(value: Any) -> bool:
        """numeric"""
        try:
            n = float(value)
        except (ValueError, TypeError):
            return False
        return mn <= n <= mx
    return _inner

def is_string(**kwargs):
    regex = None
    pattern = kwargs.get('format')
    if pattern:
        regex = re.compile(pattern)
    def _inner(value: Any) -> bool:
        """string"""
        if pattern is None:
            return type(value).__name__ == "str"
        else:
            return regex.match(str(value))
    return _inner

def is_valid_enum(**kwargs):
    symbols = kwargs.get('symbols', set())
    def _inner(value: Any) -> bool:
        """enum"""
        return value in symbols
    return _inner

def other_validator(**kwargs):
    def _inner(value: Any) -> bool:
        """other"""
        return True
    return _inner

"""
Create dictionaries to look up the type validators
"""
VALIDATORS = {
    "date": is_date,
    "nullable": is_null,
    "other": other_validator,
    "list": is_list,
    "array": is_list,
    "enum": is_valid_enum,
    "numeric": is_numeric,
    "string": is_string,
    "boolean": is_boolean,
    "cve": is_cve
}

class Schema():

    def __init__(self, definition: Union[dict, str]):
        """
        Tests a dictionary against a schema to test for conformity.
        Schema definition is similar to - but not the same as - avro schemas

        Paramaters:
            definition: dictionary or string
                A dictionary, a JSON string of a dictionary or the name of a 
                JSON file containing a schema definition
        """
        # if we have a schema as a string, load it into a dictionary
        if type(definition).__name__ == 'str':
            if os.path.exists(definition):  # type:ignore
                definition = parse(open(definition, mode='r').read())  # type:ignore
            else:
                definition = parse(definition)  # type:ignore

        try:
            # read the schema and look up the validators
            self._validators = {
                item.get('name'): self._get_validators(
                        item['type'], 
                        symbols=item.get('symbols'), 
                        min=item.get('min'),
                        max=item.get('max'),
                        format=item.get('format'))
                for item in definition.get('fields', [])  #type:ignore
            }

        except KeyError:
            raise ValueError("Invalid type specified in schema - valid types are: string, numeric, date, boolean, nullable, list, enum")
        if len(self._validators) == 0:
            raise ValueError("Invalid schema specification")


    def _get_validators(
            self,
            type_descriptor: Union[List[str], str],
            **kwargs):
        """
        For a given type definition (the ["string", "nullable"] bit), return
        the matching validator functions (the _is_x ones) as a list.
        """
        if not type(type_descriptor).__name__ == 'list':
            type_descriptor = [type_descriptor]  # type:ignore
        validators: List[Any] = []
        for descriptor in type_descriptor:
            validators.append(VALIDATORS[descriptor](**kwargs))
        return validators


    def _field_validator(
            self,
            value,
            validators: set) -> bool:
        """
        Execute a set of validator functions (the _is_x) against a value.
        Return True if any of the validators are True.
        """
        return any([True for validator in validators if validator(value)])


    def validate(self, subject: dict = {}, raise_exception=False) -> bool:
        """
        Test a dictionary against the Schema

        Parameters:
            subject: dictionary
                The dictionary to test for conformity
            raise_exception: boolean (optional, default False)
                If True, when the subject doesn't conform to the schema a
                ValidationError is raised

        Returns:
            boolean, True is subject conforms

        Raises:
            ValidationError
        """
        result = True
        self.last_error = ''
 
        for key, value in self._validators.items():
            if not self._field_validator(subject.get(key), self._validators.get(key, [other_validator])):
                result = False
                for v in value:
                    self.last_error += f"'{key}' ({subject.get(key)}) did not pass `{v.__doc__}` validator.\n"
        if raise_exception and not result:
            raise ValidationError(F"Record does not conform to schema - {self.last_error}. ")
        return result

    def __call__(self, subject: dict = {}, raise_exception=False) -> bool:
        """
        Alias for validate
        """
        return self.validate(subject=subject, raise_exception=raise_exception)

    def __str__(self):
        retval = []
        for key, value in self._validators.items():
            val = [str(v).split('.')[0].split(' ')[1] for v in value]
            val = ','.join(val)
            retval.append({"field": key, "type": val})
            
        return ascii_table(retval) 


"""
Advanced Error Trace Formatting

Adapted From:
https://github.com/willmcgugan/rich/blob/master/rich/traceback.py
Copyright 2020 Will McGugan
MIT Licence

#nodoc - don't add to the documentation wiki
"""
import os.path
import pathlib
import sys
from typing import Callable, Dict, List, Optional, Type, Any, Iterable, Tuple, TypeVar, Generator
from dataclasses import dataclass, field
from traceback import walk_tb

LINE_LENGTH = 120
BAR = ' ' * LINE_LENGTH

@dataclass
class Frame:
    filename: str
    lineno: int
    name: str
    line: str = ""
    locals: Optional[Dict[str, Any]] = None


@dataclass
class Stack:
    exc_type: str
    exc_value: str
    is_cause: bool = False
    frames: List[Frame] = field(default_factory=list)

def wrap_text(text, line_len):
    from textwrap import fill
    def _inner(text):
        for line in text.splitlines():
            yield fill(line, line_len)
    return '\n'.join(list(_inner(text)))

def bar_label(label):
    if len(label) > 0:
        center = ('**' + label + '**').center(LINE_LENGTH)
        return center.replace(' ', '-').replace('*', ' ')
    else:
        return '-' * LINE_LENGTH

def _build_error_stack():
    exc_type, exc_value, traceback = sys.exc_info()

    # if there was no error, just return
    if exc_type is None:
        return []

    stacks = []
    is_cause = False

    while True:
        stack = Stack(
                exc_type=str(exc_type.__name__),
                exc_value=str(exc_value),
                is_cause=is_cause)

        stacks.append(stack)
        append = stack.frames.append

        for frame_summary, line_no in walk_tb(traceback):
            filename = frame_summary.f_code.co_filename
            filename = os.path.abspath(filename) if filename else "?"
            frame = Frame(
                    filename=filename,
                    lineno=line_no,
                    name=frame_summary.f_code.co_name,
                    locals={key: value for key, value in frame_summary.f_locals.items()})
            append(frame)

        cause = getattr(exc_value, "__cause__", None)
        if cause and cause.__traceback__:
            exc_type = cause.__class__
            exc_value = cause
            traceback = cause.__traceback__
            if traceback:
                is_cause = True
                continue

        cause = exc_value.__context__
        if cause and cause.__traceback__ and not getattr(exc_value, "__suppress_context__", False):
            exc_type = cause.__class__
            exc_value = cause
            traceback = cause.__traceback__
            if traceback:
                is_cause = False
                continue
        # No cover, code is reached but coverage doesn't recognize it.
        break
    return stacks


def _render_locals(locals):
    if locals:
        yield bar_label('locals')
        max_label_len = max([len(key) for key, value in locals.items()])
        for key, value in locals.items():
            yield F"  {key:<{max_label_len}} : {value}"


def _read_from_code(
        filename: str,
        line: int,
        extend_by: int) -> Generator:
    try:
        with open(filename, "rt", encoding="utf-8") as code_file:
            code = code_file.read()
            
        lines = code.splitlines()
        start_line = max(line - extend_by, 0)
        end_line = min(line + extend_by, len(lines)+1)

        path = pathlib.Path(filename)

        yield bar_label(path.stem + path.suffix)
        for line_number in range(start_line, end_line):
            prefix = '>' if line_number == line else ' '
            yield F"{prefix}{line_number:4d} | {lines[line_number-1]}"
            line_number += 1
    except:
        return ""


def _render_error_stack():
    stack = _build_error_stack()
    for item in stack:
        for frame in item.frames:
            yield F"{frame.filename}:{frame.lineno} in {frame.name}()"
            if frame.filename.startswith("<"):
                continue
            yield from _render_locals(frame.locals)
            yield from _read_from_code(
                    filename=frame.filename,
                    line=frame.lineno,
                    extend_by=3)
        
            yield bar_label('')
            yield ''


def render_error_stack():
    s = [wrap_text(line, LINE_LENGTH) for line in _render_error_stack()]
    return '\n'.join(s)



"""
Ar Hap

(C) 2021 Justin Joyce.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import os
import string
from typing import Any, Union


def random_string(
        length: int = 64,
        characters: str = string.ascii_letters + string.digits) -> str:
    result_str = "".join(random_choice(characters) for _ in range(length))
    return result_str


def random_int() -> int:
    """
    Select a random integer (64bit)
    """
    return bytes_to_int(os.urandom(8))


def random_range(min: int, max: int) -> int:
    """
    Select a random integer between two numbers
    """
    return (random_int() % ((max + 1) - min)) + min


def bytes_to_int(xbytes: bytes) -> int:
    return int.from_bytes(xbytes, 'big')


def random_choice(options: Union[list, str]) -> Any:
    """
    Select an item from a list of values
    """
    r = random_range(0, len(options) - 1)
    return options[r]



"""
Google Cloud Storage Reader
"""
import io
import os
from ...data.readers.internals.base_inner_reader import BaseInnerReader
from ...utils import common, paths
from ...logging import get_logger
try:
    from google.auth.credentials import AnonymousCredentials  # type:ignore
    from google.cloud import storage  # type:ignore
except ImportError:   # pragma: no cover
    pass


class GoogleCloudStorageReader(BaseInnerReader):

    def __init__(self, project: str, **kwargs):
        super().__init__(**kwargs)
        self.project = project

    def get_blob_stream(self, object_name):
        bucket, object_path, name, extension = paths.get_parts(object_name)
        blob = get_blob(project=self.project, bucket=bucket, blob_name=object_path + name + extension)
        stream = blob.download_as_bytes()
        io_stream = io.BytesIO(stream)
        return io_stream

    def get_blobs_at_path(self, path):
        bucket, object_path, name, extension = paths.get_parts(path)

        # this means we're testing
        if os.environ.get("STORAGE_EMULATOR_HOST") is not None:
            client = storage.Client(
                credentials=AnonymousCredentials(),
                project=self.project,
            )
        else:  # pragma: no cover
            client = storage.Client(project=self.project)

        gcs_bucket = client.get_bucket(bucket)
        blobs = list(client.list_blobs(bucket_or_name=gcs_bucket, prefix=object_path))

        yield from [bucket + '/' + blob.name
                for blob in blobs
                if not blob.name.endswith('/')]


def get_blob(
        project: str,
        bucket: str,
        blob_name: str):

    # this means we're testing
    if os.environ.get("STORAGE_EMULATOR_HOST") is not None:
        client = storage.Client(
            credentials=AnonymousCredentials(),
            project=project,
        )
    else:  # pragma: no cover
        client = storage.Client(project=project)

    gcs_bucket = client.get_bucket(bucket)
    blob = gcs_bucket.get_blob(blob_name)
    return blob



"""
Base Bin

Implements common functions for each of the bins.
"""
import datetime
import abc


class BaseBin(abc.ABC):

    def __init__(self, bin_name: str):
        self.name = bin_name

    def __str__(self) -> str:
        return self.name

    @abc.abstractmethod
    def __call__(self, record: str) -> str:
        raise NotImplementedError()

    def _date_part(self):
        return datetime.date.today().strftime('%Y-%m-%d')



from ...formats.json import parse


def json_parser(ds):
    """ parse each line in the file to a dictionary """
    yield from map(parse, ds)

def pass_thru_parser(ds):
    """ just pass it through """
    yield from ds

def block_parser(ds):
    """ each blob is read as a block """
    if isinstance(ds, str):
        return ds
    yield '\n'.join([r for r in ds])


"""
from: https://stackoverflow.com/a/35804945

#nodoc - don't add to the documentation wiki
"""

import logging

def add_logging_level(level_name, level_num, method_name=None):
    """
    Comprehensively adds a new logging level to the `logging` module and the
    currently configured logging class.

    `level_name` becomes an attribute of the `logging` module with the value
    `level_num`. `methodName` becomes a convenience method for both `logging`
    itself and the class returned by `logging.getLoggerClass()` (usually just
    `logging.Logger`). If `method_name` is not specified, `levelName.lower()` is
    used.

    To avoid accidental clobberings of existing attributes, this method will
    raise an `AttributeError` if the level name is already an attribute of the
    `logging` module or if the method name is already present 

    Example:
        addLoggingLevel('TRACE', logging.DEBUG - 5)
        logging.getLogger(__name__).setLevel("TRACE")
        logging.getLogger(__name__).trace('that worked')
        logging.trace('so did this')
        logging.TRACE

    """
    if not method_name:
        method_name = level_name.lower()

    if hasattr(logging, level_name):
       raise AttributeError('{} already defined in logging module'.format(level_name))
    if hasattr(logging, method_name):
       raise AttributeError('{} already defined in logging module'.format(method_name))

    # This method was inspired by the answers to Stack Overflow post
    # http://stackoverflow.com/q/2183233/2988730, especially
    # http://stackoverflow.com/a/13638084/2988730
    def log_for_level(self, message, *args, **kwargs):
        if self.isEnabledFor(level_num):
            self._log(level_num, message, args, **kwargs)
    def log_to_root(message, *args, **kwargs):
        logging.log(level_num, message, *args, **kwargs)

    logging.addLevelName(level_num, level_name)
    setattr(logging, level_name, level_num)
    setattr(logging.getLoggerClass(), method_name, log_for_level)
    setattr(logging, method_name, log_to_root)



"""
Bloom Filter

This is a variation of the Bloom Filter; this provides a fast and memory
efficient way to tell if an item is in a list.

https://en.wikipedia.org/wiki/Bloom_filter

(C) 2021 Justin Joyce.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
try:
    from bitarray import bitarray  # type:ignore
except ImportError:
    from .bitarray import bitarray

class BloomFilter():

    __slots__ = ('filter_size', 'hash_count', 'bits')

    def __init__(
            self,
            number_of_elements: int = 50000,
            fp_rate: float = 0.05):
        """
        Bloom Filters are a probabilistic approach to tracking items in a list.
        They use an array of booleans which are set according to hashes of the
        data items. Items are considered to be in the list if the booleans at
        the hashes are set, this results in a degree of false positives. This 
        is factored into the calculation of the size of the boolean array and
        the number of hashes.

        This is used in the profiler to track unique string values without
        having to store the values or hashes of the values (minor errors 
        with this count is not expected to be a problem)

        This is expected to be used to speed-up searches for data, with
        .bloom files created to quickly determine if a file probably has the
        value being looked for (not currently implemented)
        """        
        self.filter_size = BloomFilter.get_size(number_of_elements, fp_rate)
        self.hash_count = BloomFilter.get_hash_count(self.filter_size, number_of_elements)
        self.bits = bitarray(self.filter_size, endian='big')
        self.bits.setall(0)

    @staticmethod
    def _log(x):
        return 99999999*(x**(1/99999999)-1)

    @staticmethod
    def get_size(number_of_elements, fp_rate):
        """
        Calculate the size of the bitarray

        Parameters:
            number_of_elements: integer
                The number of items expected to be stored in filter
            fp_rate: float (optional)
                False Positive rate (0 to 1), default 0.05

        Returns:
            integer
        """
        m = -(number_of_elements * BloomFilter._log(fp_rate))/(BloomFilter._log(2)**2) + 1
        return int(m)

    @staticmethod
    def get_hash_count(filter_size, number_of_elements):
        """
        Calculate the number of hashes to use to identify elements
 
        Parameters:
            filter_size: integer
                The size of the filter bit array
            number_of_elements: integer
                The number of items expected to be stored in filter

        Returns:
            integer
        """
        k = (filter_size/number_of_elements) * BloomFilter._log(2)
        return max(int(k), 2)

    def add(self, term):
        """
        Add a value to the index
        """
        import mmh3  # type:ignore
        for i in range(self.hash_count):
            h = mmh3.hash(term, seed=i) % self.filter_size
            self.bits[h] = 1

    def __contains__(self, term):
        import mmh3  # type:ignore
        for i in range(self.hash_count):
            h = mmh3.hash(term, seed=i) % self.filter_size
            if self.bits[h] == 0:
                return False
        return True

    def write_bloom_filter(self, filename: str):
        """
        Save the bloom filter to disk
        """
        with open(filename, 'wb') as fh:
            fh.write(b"MB01")
            fh.write(self.filter_size.to_bytes(4, byteorder='big'))
            fh.write(self.hash_count.to_bytes(4, byteorder='big'))
            self.bits.tofile(fh)

    @staticmethod
    def read_bloom_filter(
            filename: str):
        """
        Read a bloom filter from disk
        """
        
        def bytes_to_int(xbytes: bytes) -> int:
            return int.from_bytes(xbytes, 'big')

        bf = BloomFilter()
        
        bits = bitarray(endian='big')
        with open(filename, 'rb') as fh:
            magic_string = fh.read(4).decode()
            if magic_string != 'MB01':  # pragma: no cover
                raise IndexError(F'{filename} does appear to be a valid bloom file')
            filter_size = bytes_to_int(fh.read(4))
            hash_count  = bytes_to_int(fh.read(4)) 
            bits.fromfile(fh)

        bf.bits = bits
        bf.filter_size = filter_size
        bf.hash_count = hash_count
        return bf



import time
import datetime
from typing import Any
from dateutil import parser
from .simple_writer import SimpleWriter
from .internals.blob_writer import BlobWriter
from ..validator import Schema  # type:ignore
from ...utils import paths
from ...errors import ValidationError, InvalidDataSetError
from ...logging import get_logger
from ...data.formats import json


class BatchWriter(SimpleWriter):

    def __init__(
            self,
            *,
            dataset: str,
            format: str = 'zstd',
            date: Any = None,
            frame_id: str = None, 
            **kwargs):
        """
        The batch data writer to writes data records into blobs. Batches
        are written into timestamped folders called Partitions.

        Parameters:
            dataset: string (optional)
                The name of the dataset - this is used to map to a path
            schema: mabel.validator.Schema (optional)
                Schema used to test records for conformity, default is no 
                schema and therefore no validation
            format: string (optional)
                - jsonl: raw json lines
                - lzma: lzma compressed json lines
                - zstd: zstandard compressed json lines (default)
                - parquet: Apache Parquet
            date: date or string (optional)
                A date, a string representation of a date to use for
                creating the dataset. The default is today's date
            blob_size: integer (optional)
                The maximum size of blobs, the default is 64Mb
            inner_writer: BaseWriter (optional)
                The component used to commit data, the default writer is the
                NullWriter
            frame_id: string (optional)
            raw_path: boolean (optional)
                Don't add any date 

        Note:
            Different inner_writers may take or require additional parameters.
        """
        # call this now, we need this value before we call the base class init
        self.batch_date = self._get_writer_date(date)
        # because jobs can be split, allow a frame_id to be passed in
        if not frame_id:
            frame_id = BatchWriter.create_frame_id()

        self.dataset = dataset
        if "{date" not in self.dataset and not kwargs.get('raw_path', False):
            self.dataset += '/{datefolders}'
        self.dataset = paths.build_path(
                self.dataset + '/' + frame_id,   # type:ignore
                self.batch_date)

        kwargs['raw_path'] = True  # we've just added the dates
        kwargs['format'] = format
        kwargs['dataset'] = self.dataset

        super().__init__(**kwargs)

        # create the writer
        self.blob_writer = BlobWriter(**kwargs)

    @staticmethod
    def create_frame_id():
        return datetime.datetime.now().strftime('as_at_%Y%m%d-%H%M%S')



"""
Functions to help with handling file paths
"""
import os
import pathlib
import datetime
from uuid import uuid4

def get_parts(path_string: str):
    if not path_string:
        raise ValueError('get_parts: path_string must have a value')

    path = pathlib.PurePosixPath(path_string)
    bucket = path.parts[0]

    if len(path.parts) == 1:
        parts = 'partitions'  # type:ignore
        stem = None
        suffix = None
    elif path.suffix == "":
        parts = pathlib.PurePosixPath('/'.join(path.parts[1:-1])) / path.stem  # type:ignore
        stem = None
        suffix = None
    else:
        parts = pathlib.PurePosixPath('/'.join(path.parts[1:-1]))  # type:ignore
        stem = path.stem
        suffix = path.suffix

    return str(bucket), str(parts) + '/', stem, suffix


def build_path(path: str, date: datetime.date = None):

    if not path:
        raise ValueError('build_path: path must have a value')

    if not path[-1] in ['/']:
        # process the path
        bucket, path_string, filename, extension = get_parts(path)
        if path_string != '/':
            path_string = bucket + '/' + path_string
    else:
        path_string = path

    return date_format(path_string, date).replace('{stem}', str(uuid4()))


def date_format(path_string: str, date: datetime.date = None):

    if not date:
        date = datetime.datetime.now()

    year = date.year
    month = F"{date.month:02d}"
    day = F"{date.day:02d}"

    path_string = str(path_string)

    path_string = path_string.replace('%datefolders', f'year_{year}/month_{month}/day_{day}')
    path_string = path_string.replace('%date', F'{year}-{month}-{day}')

    path_string = path_string.replace('{datefolders}', f'year_{year}/month_{month}/day_{day}')
    path_string = path_string.replace('{date}', F'{year}-{month}-{day}')
    
    return path_string



import os
import logging
from typing import Optional
from functools import lru_cache
from .add_level import add_logging_level
from .sanitizing_log_formatter import SanitizingLogFormatter

LOG_NAME: str = "MABEL"
LOG_FORMAT: str = '%(name)s | %(levelname)-8s | %(asctime)s | %(filename)s:%(lineno)s:%(funcName)s() | %(message)s'

class LEVELS():
    """
    Proxy the Logging levels so we can just reference these without
    having to import Logging everywhere.

    DEBUG       | Information for engineers to observe inner state and flow
    INFO        | Information recording user and system events
    WARNING     | Undesirable but workable event has happened
    ERROR       | A problem has happened that stopped a minor part of the system
    AUDIT       | Information relating to proving activities happened
    ALERT       | Intervention is required
    """
    DEBUG = int(logging.DEBUG)          # 10
    INFO = int(logging.INFO)            # 20
    WARNING = int(logging.WARNING)      # 30
    ERROR = int(logging.ERROR)          # 40
    AUDIT = 80
    ALERT = 90

def set_log_name(log_name: str):
    global LOG_NAME
    LOG_NAME = log_name
    get_logger.cache_clear()

@lru_cache(1)
def get_logger() -> logging.Logger:
    """
    Use Python's native logging - we created a named logger so we can make sure
    only the logs related to our jobs are included (other modules also use the
    Python's logging module).
    """
    logger = logging.getLogger(LOG_NAME)

    # default is to log WARNING and above
    logger.setLevel(int(os.environ.get('LOGGING_LEVEL', 25)))

    # add the TRACE, AUDIT and ALERT levels to the logger
    if not hasattr(logger, 'audit'):
        add_logging_level("AUDIT", LEVELS.AUDIT)
    if not hasattr(logging, 'alert'):
        add_logging_level("ALERT", LEVELS.ALERT)

    # configure the logger
    mabel_logging_handler = logging.StreamHandler()    
    formatter = SanitizingLogFormatter(logging.Formatter(LOG_FORMAT, datefmt="%Y-%m-%dT%H:%M:%S%z"))
    mabel_logging_handler.setFormatter(formatter)
    logger.addHandler(mabel_logging_handler)

    return logger



import queue
import threading
import sys
import time
from ...formats import dictset

def threaded_reader(
        items_to_read: list, 
        reader, 
        max_threads: int = 4):
    """
    Speed up reading sets of files - such as multiple days worth of log-per-day
    files.

    Each file is read in a separate thread, up to 8 threads, reading a single
    file using this function won't show any improvement.

    NOTE:
        This compromises ordering for speed

    Parameters:
        items_to_read: list of strings:
            The name of the blobs to read
        reader: BaseReader:
            The Reader object to perform the reading operations
        max_threads: integer (optional):
            The number of threads to use to perform the read with, defaults
            to 4, is limited to 8.

    Yields:
        dictionary (or string)
    """
    thread_pool = []

    def _thread_process():
        """
        The process inside the threads.

        1) Get any files off the file queue
        2) Read the file in chunks
        3) Put a chunk onto a reply queue
        """
        try:
            source = source_queue.pop(0)
        except IndexError:
            source = None
        while source:
            source_reader = reader.get_records(source)
            for chunk in dictset.page_dictset(source_reader, 256):
                reply_queue.put(chunk)  # this will wait until there's a slot
            try:
                source = source_queue.pop(0)
            except IndexError:
                source = None


    source_queue = items_to_read.copy()

    # scale the number of threads, if we have more than the number of files
    # we're reading, will have threads that never complete
    t = min(len(source_queue), max_threads, 8)
    reply_queue: queue.Queue = queue.Queue(t * 8)

    # start the threads
    for _ in range(t):
        thread = threading.Thread(target=_thread_process)
        thread.daemon = True
        thread.start()
        thread_pool.append(thread)
        time.sleep(0.01)  # offset the start of the threads

    # when the threads are all complete and all the records have been read from
    # the reply queue, we're done
    while any([t.is_alive() for t in thread_pool]) or not(reply_queue.empty()):
        try:
            # don't wait forever
            records = reply_queue.get(timeout=10)  
            yield from records
        except queue.Empty:
            pass  #  most likely reason get being here is a race condition



from ...flows.internals.base_writer_operator import BaseWriterOperator
from ...data import BatchWriter
from ...adapters.disk import DiskWriter


class DiskBatchWriterOperator(BaseWriterOperator):

    def __init__(
            self,
            **kwargs):

        super().__init__(
                writer=BatchWriter,
                inner_writer=DiskWriter,
                **kwargs)



# python-dotenv allows us to create an environment file to store secrets. If
# there is no .env it will fail gracefully and fall back to the actual os
# environment.
try:
    from dotenv import load_dotenv   # type:ignore
    from pathlib import Path
    env_path = Path('.') / '.env'
    load_dotenv(dotenv_path=env_path)
except ImportError:
    pass

import os
from .version import __version__

if os.environ.get('RESOURCE_MONITORING', False):
    from .utils.resource_monitoring import ResourceMonitor

from .flows.flow import Flow
from .data.readers.reader import Reader
from .flows.internals.base_operator import BaseOperator
from .flows.internals.decorators import operatify



"""
Base Inner Reader
"""
import abc
import pathlib
import datetime
from io import IOBase
from typing import Iterable, Optional
from dateutil import parser
from ...formats import json
from ....utils import common, paths
from ....logging import get_logger



class BaseInnerReader(abc.ABC):

    VALID_EXTENSIONS = ('.json', '.zstd', '.lzma', '.jsonl', '.csv', '.lxml', '.parquet', '.ignore', '.profile', '.index', '.bloom')

    def _extract_date_part(self, value):
        if isinstance(value, str):
            value = parser.parse(value)
        if isinstance(value, (datetime.date, datetime.datetime)):
            return datetime.date(value.year, value.month, value.day)
        return datetime.date.today()

    def _extract_as_at(self, path):
        parts = path.split('/')
        for part in parts:
            if part.startswith('as_at_'):
                return part
        return ''

    def __init__(self, **kwargs):
        self.dataset = kwargs.get('dataset')
        if self.dataset is None:
            raise ValueError('Readers must have the `dataset` parameter set')
        if not self.dataset.endswith('/'):
            self.dataset += '/'
        if 'date' not in self.dataset and not kwargs.get('raw_path', False):
            self.dataset += '{datefolders}/'

        self.start_date = self._extract_date_part(kwargs.get('start_date'))
        self.end_date = self._extract_date_part(kwargs.get('end_date'))

        self.days_stepped_back = 0

    def step_back_a_day(self):
        """
        Steps back a day so data can be read from a previous day
        """
        self.days_stepped_back += 1
        self.start_date -= datetime.timedelta(days=1)
        self.end_date -= datetime.timedelta(days=1)
        return self.days_stepped_back

    def __del__(self):
        """
        Only here in case a helpful dev expects te base class to have it
        """
        pass

    @abc.abstractmethod
    def get_blobs_at_path(self, prefix=None) -> Iterable:
        pass 

    @abc.abstractmethod
    def get_blob_stream(self, blob: str) -> IOBase:
        """
        Return a filelike object
        """
        pass

    def get_records(
            self,
            blob_name: str,
            rows: Optional[Iterable[int]] = None) -> Iterable[str]:
        """
        Handle the different file formats.

        Handling here allows the blob stores to be pretty dumb, they
        just need to be able to store and recall blobs.

        Parameters:
            blob_name: string
                The name of the blob to read records from
            rows: list of integers (optional)
                The row numbers of the blob to retrieve records from, the
                default (None) is all of the rows. An empty set returns an
                empty Iterable from this method

        Returns:
            Iterable of the rows
        """
        # if the rows has been set but the set is empty, return empty list
        if isinstance(rows, (set,list)):
            if len(rows) == 0:
                return []
            rows = set(rows)

        stream = self.get_blob_stream(blob_name)

        if blob_name.endswith('.zstd'):
            import zstandard  # type:ignore
            with zstandard.open(stream, 'r', encoding='utf8') as file:  # type:ignore
                for index, row in enumerate(file):
                    if not rows or index in rows:
                        yield row
        elif blob_name.endswith('.lzma'):
            import lzma
            with lzma.open(stream, 'rb') as file:  # type:ignore
                for index, row in enumerate(file):
                    if not rows or index in rows:
                        yield row
        elif blob_name.endswith('.parquet'):
            try:
                import pyarrow.parquet as pq  # type:ignore
            except ImportError:
                get_logger().error("pyarrow must be installed to read parquet files")
                return []
            table = pq.read_table(stream)
            for batch in table.to_batches():
                dict_batch = batch.to_pydict()
                for index in range(len(batch)):
                    if not rows or index in rows:
                        yield json.serialize({k:v[index] for k,v in dict_batch.items()})  # type:ignore
        else:  # assume text in lines format
            text = stream.read().decode('utf8')  # type:ignore
            lines = text.splitlines()
            for index, row in enumerate(lines):
                if not rows or index in rows:
                    if len(row) > 0:
                        yield row


    def get_list_of_blobs(self):

        blobs = []
        for cycle_date in common.date_range(self.start_date, self.end_date):
            # build the path name
            cycle_path = pathlib.Path(paths.build_path(path=self.dataset, date=cycle_date))
            blobs += list(self.get_blobs_at_path(path=cycle_path))

        # remove any BACKOUT data
        blobs = [blob for blob in blobs if 'BACKOUT' not in blob]

        # work out if there's an as_at part
        as_ats = { self._extract_as_at(blob) for blob in blobs if 'as_at_' in blob }
        if as_ats:
            as_ats = sorted(as_ats)
            as_at = as_ats.pop()
            # the .ignore file means the frame shouldn't be used 
            while any([blob for blob in blobs if as_at + '/.ignore' in blob]):
                get_logger().debug(F".ignore file found in frame {as_at}, ignoring")
                if len(as_ats) > 0:
                    as_at = as_ats.pop()
                else:
                    return []
            get_logger().debug(F"Reading from DataSet frame {as_at}")
            blobs = [blob for blob in blobs if as_at in blob]

        return blobs



#no-maintain-checks
"""
This module is dervied from:

Cryno: Data Profiling Library
https://github.com/joocer/cryno

(C) 2021 Justin Joyce.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from ..flows.internals.base_operator import BaseOperator
from ..data.formats import json
from ..data.formats.dictset.display import draw_histogram_bins
from ..index.bloom_filter import BloomFilter


class ProfileDataOperator(BaseOperator):

    def __init__(self, 
            schema = None,
            inner_writer = None,
            **kwargs):
        """
        ProfileData Operator is designed to profile streaming data this is done
        with some approximation, especially when summarizations would require
        all of the values such as counting unique values or binning values.

        It starts summarizing from 10,000 values. So small datasets should be
        more accurate.

        Some measures will be pretty much useless the closer it gets to
        5,000,000 records.
        """
        super().__init__(**kwargs)

        self.inner_writer = inner_writer
        self.fields = schema._validators

        self.summary = {}
        for field in self.fields:
            self.summary[field] = {
                "nulls": 0,
                "items": 0,
                "type": ProfileDataOperator.get_type(schema._validators[field])
            }

    def execute(self, data, context):

        for field in self.fields:
            field_summary = self.summary[field]
            field_summary['items'] += 1
            field_type = field_summary['type']
            field_value = data.get(field)
            if field_value is None:
                field_summary['nulls'] += 1

            elif field_type == 'numeric':
                field_summary = ProfileDataOperator.profile_numeric_data(field_summary, field_value)
                    
            elif field_type == 'date':
                # convert to epoch seconds and then just treat as a number
                from dateutil import parser
                field_value = int(parser.isoparse(field_value).timestamp())
                field_summary = ProfileDataOperator.profile_numeric_data(field_summary, field_value)

            elif field_type == 'string':
                field_summary = ProfileDataOperator.profile_string_data(field_summary, field_value)

            elif field_type == 'enum':
                """
                Profiler for enums (including boolean), just counts the
                number of values in each category
                """
                if field_summary.get('values') is None:
                    field_summary['values'] = {}
                field_summary['values'][F"{field_value}"] = field_summary['values'].get(F"{field_value}", 0) + 1

        return data, context

    def finalize(self, context: dict = None):

        if not context:
            context = {}

        for k in [k for k in self.summary if self.summary[k]['type'] == 'string']:

            if len(self.summary[k]['unique_values']) < 10000:
                # get rid of the unique values list
                self.summary[k]['unique_values'] = len(self.summary[k]['unique_values'])
                del self.summary[k]['approx_unique_values']
            else:
                del self.summary[k]['unique_values']

            del self.summary[k]['bloom_filter']


        # rebin all of the binned data into a set of 10 bins
        new_bins = [(k, ProfileDataOperator.redistribute_bins(self.summary[k]['bins'], 10)) for k in self.summary if 'bins' in self.summary[k]]
        for k, bins in new_bins:
            self.summary[k]['bins'] = bins

        # convert date fields back to dates
        date_fields = [k for k in self.summary if self.summary[k]['type'] == 'date']
        for date_field in date_fields:
            self.summary[date_field]['min'] = ProfileDataOperator.date_from_epoch(self.summary[date_field]['min'])
            self.summary[date_field]['max'] = ProfileDataOperator.date_from_epoch(self.summary[date_field]['max'])
            new_bins = {}  # type:ignore
            for bound in self.summary[date_field]['bins']:
                bottom, top = bound
                bottom = ProfileDataOperator.date_from_epoch(bottom)
                top = ProfileDataOperator.date_from_epoch(top)
                new_bins[(bottom, top)] = self.summary[date_field]['bins'][bound]  # type:ignore
            self.summary[date_field]['bins'] = new_bins
            del self.summary[date_field]['mean']
            del self.summary[date_field]['cumsum']

        # change the bin names to strings from tuples
        for k in [k for k in self.summary if 'bins' in self.summary[k]]:
            new_bins = {}  # type:ignore
            old_bins = self.summary[k]['bins']
            for ranges, v in old_bins.items():
                bottom, top = ranges
                label = F"{ProfileDataOperator.short_form(bottom)} to {ProfileDataOperator.short_form(top)}"
                new_bins[label] = v  # type:ignore
            self.summary[k]['bins'] = new_bins

        # put the profile into the context so it gets passed along
        context['mabel:profile'] = self.summary
        return context

    def __str__(self):
        def _inner():
            for field in self.fields:
                field_summary = self.summary[field]
                if field_summary['nulls'] == field_summary['items']:
                    yield F"[---] {field:20} [count] {field_summary['items']} EMPTY"
                elif field_summary['type'] == 'numeric':
                    yield F"[num] {field:20} [count] {field_summary['items']} [empty] {(field_summary['nulls'] / field_summary['items']):.1%} [range] {ProfileDataOperator.short_form(field_summary['min'])} to {ProfileDataOperator.short_form(field_summary['max'])} [mean] {ProfileDataOperator.short_form(field_summary['mean'])} >{draw_histogram_bins([v for k,v in field_summary['bins'].items()])}<"
                elif field_summary['type'] == 'date':
                    yield F"[dte] {field:20} [count] {field_summary['items']} [empty] {(field_summary['nulls'] / field_summary['items']):.1%} [range] {field_summary['min']} to {field_summary['max']} >{draw_histogram_bins([v for k,v in field_summary['bins'].items()])}<"
                elif field_summary['type'] == 'other':
                    yield F"[oth] {field:20} [count] {field_summary['items']} [empty] {(field_summary['nulls'] / field_summary['items']):.1%}"
                elif field_summary['type'] == 'string':
                    yield F"[str] {field:20} [count] {field_summary['items']} [empty] {(field_summary['nulls'] / field_summary['items']):.1%} [unique] {field_summary['unique_values']} [longest] {field_summary['max_length']} [shortest] {field_summary['min_length']}"
                elif field_summary['type'] == 'enum':
                    yield F"[enm] {field:20} [count] {field_summary['items']} [empty] {(field_summary['nulls'] / field_summary['items']):.1%} {ProfileDataOperator.enum_summary(field_summary['values'])}"
        return '\n'.join(_inner())

    def __repr__(self):
        return json.serialize(self.summary, indent=True)

    @staticmethod
    def redistribute_bins(bins, number_of_bins=100):
        """
        Redistribute binned data into fewer bins

        This results in items being put into incorrect bins, especially
        when the data is ordered.
        """
        mn = min([l for l,h in bins])
        mx = max([h for l,h in bins])
        
        bin_size = ((mx - mn) + 1) // number_of_bins
        
        new_bins = {}
        for counter in range(number_of_bins - 1):
            new_bins[(mn + (bin_size * counter), mn + (bin_size * (counter + 1)) - 1)] = 0
        new_bins[(mn + (bin_size * (counter + 1)),mx)] = 0
            
        for old_bounds in bins:
            old_lower, old_upper = old_bounds
            old_mid = (old_lower + old_upper) // 2
            binned = False
            for new_bounds in new_bins:
                new_lower, new_upper = new_bounds
                if old_mid >= new_lower and old_mid <= new_upper:
                    new_bins[new_bounds] += bins[old_bounds]
                    binned = True
                    break
            if binned == False:
                print('miss', old_mid, mn, mx)

        return new_bins

    @staticmethod
    def get_type(validators):
        """
        Determine the data type so we can use the correct profiler
        """
        try:
            val = [str(v).split('.')[0].split(' ')[1] for v in validators]
            val = [v for v in val if not v in ['is_null', 'is_other']]
            val = val.pop()
        except (KeyError, ValueError, IndexError):
            val = "other"
        
        if val in ['is_numeric']:
            return "numeric"
        if val in ['is_string', 'is_cve']:
            return "string"
        if val in ['is_valid_enum', 'is_boolean']:
            return "enum"
        if val in ['is_date']:
            return "date"

        return "other"

    @staticmethod
    def date_from_epoch(seconds, form='%Y-%m-%d %H:%M:%S'):
        """
        Convert from milliseconds since epoch to a date string
        """
        import datetime
        return datetime.datetime.fromtimestamp(seconds).strftime(form)

    @staticmethod
    def profile_numeric_data(collector, value):
        """
        Profiler for numeric data, it collects:

        - minimim value
        - maximum value
        - cummulative sum
        - mean
        - approximate distribution

        This method take signle value at a time to build the profile
        """
        if collector.get('max', value) <= value:
            collector['max'] = value
        if collector.get('min', value) >= value:
            collector['min'] = value
        collector['cumsum'] = collector.get('cumsum', 0) + value
        collector['mean'] = collector['cumsum'] / collector['items']
            
        if collector.get('bins') is None:
            collector['bins'] = {}
        binned = False
        for bounds in collector['bins']:
            bottom, top = bounds
            if value >= bottom and value <= top:
                collector['bins'][bounds] += 1
                binned = True
        if not binned:
            collector['bins'][(value, value)] = 1
            
        if len(collector['bins']) > 10007:   # first prime over 10000
            collector['bins'] = ProfileDataOperator.redistribute_bins(collector['bins'], 101)  # first prime over 100

        return collector

    @staticmethod
    def profile_string_data(collector, value):
        """
        Profiler for string data, it collects:

        - minimim length
        - maximum length
        - unique values (capped upper limit)

        This method take signle value at a time to build the profile
        """
        if collector.get('max_length', len(value)) <= len(value):
            collector['max_length'] = len(value)
        if collector.get('min_length', len(value)) >= len(value):
            collector['min_length'] = len(value)
        if collector.get('bloom_filter') is None:
            collector['unique_values'] = []
            collector['bloom_filter'] = BloomFilter(number_of_elements=5000000, fp_rate=0.01)
            collector['approx_unique_values'] = 0

        # if we have only a few values, count them all
        if len(collector['unique_values']) < 10000:
            hashed = hash(value)
            if not hashed in collector['unique_values']:
                collector['unique_values'].append(hashed)

        # when we have a lot of values, use a bloom filter
        if value not in collector['bloom_filter']:
            collector['bloom_filter'].add(value)
            collector['approx_unique_values'] += 1

        return collector

    @staticmethod
    def short_form(num):
        if not isinstance(num, (int, float)):
            return num
        if isinstance(num, int) and abs(num) < 10000000:
            return str(num)
        return F"{num:.2f}"

    @staticmethod
    def enum_summary(dic):
        s = {k:v for k,v in sorted(dic.items(), key=lambda item: item[1], reverse=True)}
        cumsum = sum([v for k,v in dic.items()])
        eliminated = 0
        result = ''
        for index, item in enumerate(s):
            if index == 2:
                break
            result += F"`{item}`: {(s[item] / cumsum):.1%} "
            eliminated += s[item]
        if eliminated < cumsum:
            result += F"[other]: {(cumsum - eliminated) / cumsum:.1%}"
        return result



from .base_exception import BaseException
from .render_error_stack import render_error_stack
from .validation_error import ValidationError
from .invalid_combination_error import InvalidCombinationError
from .invalid_data_set_error import InvalidDataSetError
from .integrity_error import IntegrityError
from .invalid_reader_configuration import InvalidReaderConfigError
from .flow_error import FlowError
from .invalid_syntax import InvalidSyntaxError



import time
import datetime
from typing import Any
from dateutil import parser
from .internals.blob_writer import BlobWriter
from ..validator import Schema  # type:ignore
from ...utils import paths
from ...errors import ValidationError, InvalidDataSetError
from ...logging import get_logger
from ...data.formats import json


class SimpleWriter():

    def _get_writer_date(self, date):
        # default to today if not given a date
        batch_date = datetime.datetime.now()
        if isinstance(date, datetime.date):
            batch_date = date  # type:ignore
        if isinstance(date, str):
            batch_date = parser.parse(date)
        return batch_date


    def __init__(
            self,
            *,
            dataset: str,
            schema: Schema = None,
            format: str = 'zstd',
            date: Any = None,
            **kwargs):
        """
        Simple Writer provides a basic writer capability.
        """
        if 'BACKOUT' in dataset:
            InvalidDataSetError('BACKOUT is a reserved word and cannot be used in Dataset names')
        if dataset.endswith('/'):
            InvalidDataSetError('Dataset names cannot end with /')
        if '{' in dataset or '}' in dataset:
            InvalidDataSetError('Dataset names cannot contain { or }')
        if '%' in dataset:
            InvalidDataSetError('Dataset names cannot contain %')

        self.schema = schema
        self.finalized = False
        self.batch_date = self._get_writer_date(date)

        self.dataset = dataset
        if "{date" not in self.dataset and not kwargs.get('raw_path', False):
            self.dataset += '/{datefolders}'
        self.dataset = paths.build_path(self.dataset, self.batch_date)

        # add the values to kwargs
        kwargs['raw_path'] = True  # we've just added the dates
        kwargs['format'] = format
        kwargs['dataset'] = self.dataset

        arg_dict = kwargs.copy()
        arg_dict['dataset'] = F'{self.dataset}'
        arg_dict['inner_writer'] = F"{arg_dict.get('inner_writer', {}).__name__}"  # type:ignore
        get_logger().debug(json.serialize(arg_dict))

        # create the writer
        self.blob_writer = BlobWriter(**kwargs)


    def append(self, record: dict = {}):
        """
        Append a new record to the Writer

        Parameters:
            record: dictionary
                The record to append to the Writer

        Returns:
            integer
                The number of records in the current blob
        """
        # Check the new record conforms to the schema before continuing
        if self.schema and not self.schema.validate(subject=record, raise_exception=False):
            raise ValidationError(F'Schema Validation Failed ({self.schema.last_error})')

        self.blob_writer.append(record)

    def __del__(self):
        if hasattr(self, 'finalized') and not self.finalized:
            get_logger().error(F"{type(self).__name__} has not been finalized - data may be lost, make sure you're call .finalize()")

    def finalize(self):
        self.finalized = True
        try:
            self.blob_writer.commit()
        except Exception as e:
            get_logger().error(F"{type(self).__name__} failed to close pool: {type(e).__name__} - {e}")



from typing import Callable
from ....logging import get_logger

class Groups():
    """
    Group By functionality for Iterables of Dictionaries
    
    Parameters:
        dictset: Iterable of dictionaries:
            The dataset to perform the Group By on
        column: string:
            The name of the field to group by
    
    Returns:
        Groups

    Warning:
        The 'Groups' object holds the entire dataset in memory so is unsuitable
        for large datasets.
    """
    __slots__ = ('_groups')

    def __init__(self, dictset, column):
        groups = {}
        for item in dictset:
            my_item = item.copy()
            key = my_item.get(column)
            if groups.get(key) is None:
                groups[my_item.get(column)] = []
            del my_item[column]
            groups[key].append(my_item)
        self._groups = groups

    def count(self, group=None):
        """
        Count the number of items in groups
        
        Parameters:
            group: string (optional)
                If provided, return the count of just this group

        Returns:
            if a group is provided, return an integer
            if no group is provided, return a dictionary
        """
        if group is None:
            return {x:len(y) for x,y in self._groups.items()}
        else:
            try:
                return [len(y) for x,y in self._groups.items() if x == group].pop()
            except IndexError:
                return 0

    def aggregate(self, column, method):
        """
        Applies an aggregation function by group.
        
        Parameters:
            column: string
                The name of the field to aggregate on
            method: callable
                The function to aggregate with
        
        Returns:
            dictionary

        Examples:
            maxes = grouped.aggregate('age', max)
            means = grouped.aggregate('age', maths.mean)
        """
        response = {}
        for key, items in self._groups.items():
            values = [item.get(column) for item in items if item.get(column) is not None]
            response[key] = method(values)
        return response

    def apply(self, method: Callable):
        """
        Apply a function to all groups

        Parameters:
            method: callable
                The function to apply to the groups

        Returns:
            dictionary
        """
        return {key:method(items) for key, items in self._groups.items()}
            
    def __len__(self):
        """
        Returns the number of groups in the set.
        """
        return len(self._groups)

    def __repr__(self):
        """
        Returns the group names
        """
        return str(list(self._groups.keys()))



# type:ignore
import io
import glob
import datetime
from typing import Iterable, Tuple, Optional, List
from ...data.readers.internals.base_inner_reader import BaseInnerReader
from ...utils import paths, common
from ...logging import get_logger
from ...errors import InvalidReaderConfigError
try:
    import paho.mqtt.client as mqtt
except ImportError:  # pragma: no cover
    pass


# pip install paho-mqtt

class MqttReader(BaseInnerReader):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.dataset = kwargs.get('dataset')
        self.host = kwargs.get('host')
        if not self.host:
            raise InvalidReaderConfigError('MqttReader requires `host` to be set.')
        self.port = int(kwargs.get('port', 1883))

        self.messages = []


    def on_connect(self, client, userdata, flags, rc):
        get_logger().debug(F"Connected to mqtt server {self.host} with result code {rc}")
        client.subscribe(self.dataset)

    def on_message(self, client, userdata, msg):
        self.messages.append(msg.payload)

    def get_list_of_blobs(self):
        """
        Override this method to just return the queue we're interested in
        """
        return [self.dataset]

    def get_blobs_at_path(self, path):
        # not used but must be present
        pass

    def get_blob_stream(self, blob_name:str) -> io.IOBase:
        # not used but must be present
        pass

    def get_records(
            self,
            blob_name: str,
            rows: Optional[Iterable[int]] = None) -> Iterable[str]:
        """
        Override this method to just return records as they appear
        on the queue
        """
        self.client = mqtt.Client()
        self.client.on_connect = self.on_connect
        self.client.on_message = self.on_message
        self.client.connect(self.host, self.port, 60)

        while True:
            self.client.loop(timeout=1.0)
            if self.messages:
                yield from self.messages
            self.messages.clear()

    def __del__(self):
        self.client.loop_stop()
